from pelops.mythreading import LoggerThread
import queue
from pelops.logging import mylogger


class Worker:
    """
    Creates n worker threads that wait for functions to be executed. Thus, a simple way of parallelizing
    processing of many different tasks. A single queue is filled and the worker threads take the oldest
    entry and executes them.
    """

    _list = None  # list containing the worker thread references.
    queue = None  # queue of to be processed tasks
    _number = None  # number of worker threads that should be created.
    _logger = None  # logger instance

    def __init__(self, number_worker, parent_logger):
        """
        Constructor.

        :param number_worker: number of worker threads that should be created.
        :param parent_logger: logger instance from the parent. a child will be spawned
        """

        self._logger = mylogger.get_child(parent_logger, self.__class__.__name__)
        self._logger.info("Worker.__init__ - initializing")
        self._logger.debug("Worker.__init__ - config: number_worker={}.".format(number_worker))
        self.queue = queue.Queue()
        self._list = []
        self._number = int(number_worker)

    def _worker_thread(self):
        self._logger.info("Worker._worker_thread - started worker")
        while True:
            item = self.queue.get()
            self._logger.info("Worker._worker_thread - worker received item '{}'.".format(item))
            if item is None:
                self._logger.info("Worker._worker_thread - worker received stop signal")
                break
            func, parameter = item
            func(parameter)
            self.queue.task_done()
        self._logger.info("Worker._worker_thread - stopped worker")

    def start(self):
        """
        Starts all worker threads.
        """

        self._logger.info("Worker.start - starting {} worker.".format(self._number))
        for i in range(self._number):
            w = LoggerThread(target=self._worker_thread, name="worker_{}".format(self._number), logger=self._logger)
            w.start()
            self._list.append(w)
        self._logger.info("Worker.start - {} worker started.".format(len(self._list)))

    def stop(self):
        """
        Stops all worker threads.
        """

        self._logger.info("Worker.stop - stopping {} worker.".format(len(self._list)))
        for i in range(len(self._list)):
            self.queue.put(None)
        for t in self._list:
            t.join()
        self._list = []
        self._logger.info("Worker.stop - worker stopped.")



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops import myconfigtools
from pelops.logging import mylogger
from alcathous.worker import Worker
import time


class TestWorker(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        config = myconfigtools.read_config(
            os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        cls.logger = mylogger.create_logger(config["logger"], __name__)
        cls.logger.info("TestWorker - start")
        cls.config = config["data-preparation"]

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("TestWorker - stop")

    def test_worker0_init(self):
        worker = Worker(self.config["number_worker"], self.logger)
        self.assertEqual(worker._number, 2)
        self.assertIsNotNone(worker.queue)
        self.assertIsNotNone(worker._logger)
        self.assertIsNotNone(worker._list)
        self.assertEqual(len(worker._list), 0)

    def test_worker1_start_stop(self):
        worker = Worker(self.config["number_worker"], self.logger)
        worker.start()
        self.assertEqual(len(worker._list), 2)
        worker.stop()
        self.assertEqual(len(worker._list), 0)

    def test_worker2_queue(self):
        worker = Worker(self.config["number_worker"], self.logger)
        worker.start()
        start_time = time.time()
        for i in range(4):
            worker.queue.put((time.sleep, 1))
        worker.stop()  # must be called before taking time
        stop_time = time.time()
        diff_time = stop_time - start_time
        self.assertLess(diff_time, 3)  # two workers are processing 4*sleep(1) - it should be done in close to 2 ...
        self.assertGreater(diff_time, 2)  # it must take at least 2 seconds ...


if __name__ == '__main__':
    unittest.main()


from alcathous.algorithms.abstractalgorithm import AbstractAlgorithm


class Count(AbstractAlgorithm):
    """
    Count how many valid data points are within the give time window.
    """

    def _process(self, time_from, time_to):
        self._logger.info("{} - average for '{} s' to '{} s'.".format(self._topic, time_from, time_to))
        sum_values = 0
        count = 0
        with self._lock_data_set:
            self._logger.info("{} - acquired lock".format(self._topic))
            for timestamp, value in reversed(self._data_set.items()):
                self._logger.info("{} - timestamp '{} s'; value '{}'.".format(self._topic, timestamp, value))
                if time_from > timestamp:
                    self._logger.info("{} - time_from > timestamp".format(self._topic))
                    break
                elif timestamp <= time_to:
                    count = count + 1
        self._logger.info("{} - sum: {}; count: {}.".format(self._topic, sum_values, count))
        return count

    def execution_points_estimation(self):
        return int(self._time_window*0.9)



from alcathous.algorithms.abstractalgorithm import AbstractAlgorithm


class Minimum(AbstractAlgorithm):
    """
    The minimum value of all valid data points within the time window.
    """

    def _process(self, time_from, time_to):
        self._logger.info("{} - average for '{} s' to '{} s'.".format(self._topic, time_from, time_to))
        sum_values = 0
        count = 0
        minimum = None
        with self._lock_data_set:
            self._logger.info("{} - acquired lock".format(self._topic))
            for timestamp, value in reversed(self._data_set.items()):
                self._logger.info("{} - timestamp '{} s'; value '{}'.".format(self._topic, timestamp, value))
                if time_from > timestamp:
                    self._logger.info("{} - time_from > timestamp".format(self._topic))
                    break
                elif timestamp <= time_to:
                    try:
                        minimum = min(minimum, value)
                    except TypeError:
                        minimum = value
                    count = count + 1
        self._logger.info("{} - sum: {}; count: {}.".format(self._topic, sum_values, count))
        if count == 0:
            raise ValueError
        return minimum

    def execution_points_estimation(self):
        return int(self._time_window)
